import numpy as np
from scipy.sparse import csr_matrix
from scipy.stats import norm

from anndata import AnnData
from pandas import DataFrame
from typing import Optional

from tqdm import tqdm

from liana.method.sp._SpatialMethod import SpatialMethod
from liana.method._global_lr_pipe import _global_lr_pipe
from liana.method.sp._spatial_utils import _local_to_dataframe, _local_permutation_pvals


class SpatialDM(SpatialMethod):
    def __init__(self, _method, _complex_cols, _obsm_keys):
        super().__init__(method_name=_method.method_name,
                         key_cols=_method.key_cols,
                         reference=_method.reference,
                         )

        self.complex_cols = _complex_cols
        self.obsm_keys = _obsm_keys
        self._method = _method

    def __call__(self,
                 adata: AnnData,
                 resource_name: str = 'consensus',
                 expr_prop: float = 0.05,
                 pvalue_method: str = 'analytical',
                 n_perm: int = 1000,
                 positive_only: bool = True,
                 use_raw: Optional[bool] = True,
                 layer: Optional[str] = None,
                 verbose: Optional[bool] = False,
                 seed: int = 1337,
                 resource: Optional[DataFrame] = None,
                 inplace=True
                 ):
        """
        Parameters
        ----------
        adata
            Annotated data object.
        resource_name
            Name of the resource to be loaded and use for ligand-receptor inference.
        expr_prop
            Minimum expression proportion for the ligands/receptors (and their subunits).
             Set to `0` to return unfiltered results.
        pvalue_method
            Method to obtain P-values: One out of ['permutation', 'analytical'];
            'analytical' by default.
        n_perm
            Number of permutations to be performed if `pvalue_method`=='permutation'
        positive_only
            Whether to calculate p-values only for positive correlations. `True` by default.
        use_raw
            Use raw attribute of adata if present.
        layer
            Layer in anndata.AnnData.layers to use. If None, use anndata.AnnData.X.
        verbose
            Verbosity flag
        seed
            Random seed for reproducibility.
        resource
            Parameter to enable external resources to be passed. Expects a pandas dataframe
            with [`ligand`, `receptor`] columns. None by default. If provided will overrule
            the resource requested via `resource_name`
        inplace
            If true return `DataFrame` with results, else assign to `.uns`.

        Returns
        -------
        If ``inplace = False``, returns:
        - 1) a `DataFrame` with ligand-receptor correlations for the whole slide (global)
        - 2) a `DataFrame` with ligand-receptor Moran's I for each spot
        - 3) a `DataFrame` with ligand-receptor correlations p-values for each spot
        Otherwise, modifies the ``adata`` object with the following keys:
        - :attr:`anndata.AnnData.uns` ``['global_res']`` with `1)`
        - :attr:`anndata.AnnData.obsm` ``['local_r']`` with  `2)`
        - :attr:`anndata.AnnData.obsm` ``['local_pvals']`` with  `3)`

        """
        if pvalue_method not in ['analytical', 'permutation']:
            raise ValueError('pvalue_method must be one of [analytical, permutation]')

        temp, lr_res, ligand_pos, receptor_pos = _global_lr_pipe(adata=adata,
                                                                 resource_name=resource_name,
                                                                 resource=resource,
                                                                 expr_prop=expr_prop,
                                                                 use_raw=use_raw,
                                                                 layer=layer,
                                                                 verbose=verbose,
                                                                 _key_cols=self.key_cols,
                                                                 _complex_cols=self.complex_cols,
                                                                 _obms_keys=self.obsm_keys
                                                                 )

        # n / sum(W) for Moran's I
        norm_factor = temp.obsm['proximity'].shape[0] / temp.obsm['proximity'].sum()
        dist = csr_matrix(norm_factor * temp.obsm['proximity'])

        # we use the same gene expression matrix for both x and y
        lr_res['global_r'], lr_res['global_pvals'] = \
            _global_spatialdm(x_mat=temp.X,
                              y_mat=temp.X,
                              x_pos=ligand_pos,
                              y_pos=receptor_pos,
                              xy_dataframe=lr_res,
                              dist=dist,
                              seed=seed,
                              n_perm=n_perm,
                              pvalue_method=pvalue_method,
                              positive_only=positive_only
                              )
        local_r, local_pvals = _local_spatialdm(x_mat=temp.X,
                                                y_mat=temp.X,
                                                x_pos=ligand_pos,
                                                y_pos=receptor_pos,
                                                xy_dataframe=lr_res,
                                                dist=dist,  # TODO msq?
                                                seed=seed,
                                                n_perm=n_perm,
                                                pvalue_method=pvalue_method,
                                                positive_only=positive_only
                                                )

        # convert to dataframes
        local_r = _local_to_dataframe(array=local_r,
                                      idx=temp.obs.index,
                                      columns=lr_res.interaction)
        local_pvals = _local_to_dataframe(array=local_pvals,
                                          idx=temp.obs.index,
                                          columns=lr_res.interaction)

        if inplace:
            adata.uns['global_res'] = lr_res
            adata.obsm['local_r'] = local_r
            adata.obsm['local_pvals'] = local_pvals

        return None if inplace else (lr_res, local_r, local_pvals)


def _global_spatialdm(x_mat,
                      y_mat,
                      x_pos,
                      y_pos,
                      xy_dataframe,
                      dist,
                      seed,
                      n_perm,
                      pvalue_method,
                      positive_only,
                      x_key = 'ligand',
                      y_key = 'receptor'):
    """
    Global Moran's Bivariate I as implemented in SpatialDM

    Parameters
    ----------
    x_mat
        Gene expression matrix for entity x (e.g. ligand)
    y_mat
        Gene expression matrix for entity y (e.g. receptor)
    x_pos
        Index positions of entity x (e.g. ligand) in `mat`
    y_pos
        Index positions of entity y (e.g. receptor) in `mat`
    xy_dataframe
        a dataframe with x,y relationships to be estimated, for example `lr_res`.
    dist
        proximity weight matrix, obtained e.g. via `liana.method.get_spatial_proximity`.
        Note that for spatialDM/Morans'I `dist` has to be weighed by n / sum(W).
    seed
        Reproducibility seed
    n_perm
        Number of permutatins to perform (if `pvalue_method`=='permutation')
    pvalue_method
        Method to estimate pseudo p-value, must be in ['permutation', 'analytical']
    positive_only
        Whether to return only p-values for positive spatial correlations.
        By default, `True`.

    Returns
    -------
    Tupple of 2 1D Numpy arrays of size xy_dataframe.shape[1],
    or in other words calculates global_I and global_pval for
    each interaction in `xy_dataframe`

    """

    # normalize matrices
    x_mat = _standardize_matrix(x_mat, local=False)
    y_mat = _standardize_matrix(y_mat, local=False)

    # convert to spot_n x lr_n matrices
    x_mat = _get_ordered_matrix(mat=x_mat,
                                pos=x_pos,
                                order=xy_dataframe[x_key])
    y_mat = _get_ordered_matrix(mat=y_mat,
                                pos=y_pos,
                                order=xy_dataframe[y_key])

    # Get global r
    global_r = ((x_mat @ dist) * y_mat).sum(axis=1)

    # calc p-values
    if pvalue_method == 'permutation':
        global_pvals = _global_permutation_pvals(x_mat=x_mat,
                                                 y_mat=y_mat,
                                                 dist=dist,
                                                 global_r=global_r,
                                                 n_perm=n_perm,
                                                 positive_only=positive_only,
                                                 seed=seed
                                                 )
    elif pvalue_method == 'analytical':
        global_pvals = _global_zscore_pvals(dist=dist,
                                            global_r=global_r,
                                            positive_only=positive_only)

    return global_r, global_pvals



def _global_permutation_pvals(x_mat, y_mat, dist, global_r, n_perm, positive_only, seed):
    """
    Calculate permutation pvals

    Parameters
    ----------
    x_mat
        Matrix with x variables
    y_mat
        Matrix with y variables
    dist
        Proximity weights 2D array
    global_r
        Global Moran's I, 1D array
    n_perm
        Number of permutations
    positive_only
        Whether to mask negative p-values
    seed
        Reproducibility seed

    Returns
    -------
    1D array with same size as global_r

    """
    assert isinstance(dist, csr_matrix)
    rng = np.random.default_rng(seed)

    # initialize mat /w n_perm * number of X->Y
    idx = x_mat.shape[1]

    # permutation mat /w n_perms x LR_n
    perm_mat = np.zeros((n_perm, global_r.shape[0]))

    for perm in tqdm(range(n_perm)):
        _idx = rng.permutation(idx)
        perm_mat[perm, :] = ((x_mat[:, _idx] @ dist) * y_mat).sum(axis=1) # flipped x_mat

    if positive_only:
        global_pvals = 1 - (global_r > perm_mat).sum(axis=0) / n_perm
    else:
        # TODO Proof this makes sense
        global_pvals = 2 * (1 - (np.abs(global_r) > np.abs(perm_mat)).sum(axis=0) / n_perm)

    return global_pvals



def _local_spatialdm(x_mat,
                     y_mat,
                     x_pos,
                     y_pos,
                     xy_dataframe,
                     dist,
                     n_perm,
                     seed,
                     pvalue_method,
                     positive_only,
                     x_key = 'ligand',
                     y_key = 'receptor'
                     ):
    """
    Local Moran's Bivariate I as implemented in SpatialDM

    Parameters
    ----------
    x_mat
        Matrix with x variables
    y_mat
        Matrix with y variables
    x_pos
        Index positions of entity x (e.g. ligand) in `mat`
    y_pos
        Index positions of entity y (e.g. receptor) in `mat`
    xy_dataframe
        a dataframe with x,y relationships to be estimated, for example `lr_res`.
    dist
        proximity weight matrix, obtained e.g. via `liana.method.get_spatial_proximity`.
        Note that for spatialDM/Morans'I `dist` has to be weighed by n / sum(W).
    seed
        Reproducibility seed
    n_perm
        Number of permutatins to perform (if `pvalue_method`=='permutation')
    pvalue_method
        Method to estimate pseudo p-value, must be in ['permutation', 'analytical']
    positive_only
        Whether to return only p-values for positive spatial correlations.
        By default, `True`.

    Returns
    -------
        Tupple of two 2D Numpy arrays of size (n_spots, n_xy),
         or in other words calculates local_I and local_pval for
         each interaction in `xy_dataframe` and each sample in mat
    """
    x_mat = _standardize_matrix(x_mat, local=True)
    y_mat = _standardize_matrix(y_mat, local=True)

    ligand_mat = _get_ordered_matrix(mat=x_mat,
                                     pos=x_pos,
                                     order=xy_dataframe[x_key])
    receptor_mat = _get_ordered_matrix(mat=y_mat,
                                       pos=y_pos,
                                       order=xy_dataframe[y_key])
    ligand_mat, receptor_mat = ligand_mat.T, receptor_mat.T
    local_r = _calculate_local_moransI(ligand_mat, receptor_mat, dist)

    if pvalue_method == 'permutation':
        local_pvals = _local_permutation_pvals(x_mat=ligand_mat,
                                               y_mat=receptor_mat,
                                               dist=dist,
                                               local_truth=local_r,
                                               local_fun=_calculate_local_moransI,
                                               n_perm=n_perm,
                                               seed=seed,
                                               positive_only=positive_only
                                               )
    elif pvalue_method == 'analytical':
        local_pvals = _local_zscore_pvals(x_mat=ligand_mat,
                                          y_mat=receptor_mat,
                                          local_r=local_r,
                                          dist=dist,
                                          positive_only=positive_only)

    return local_r.T, local_pvals.T


def _standardize_matrix(mat, local=True):
    mat = np.array(mat - np.array(mat.mean(axis=0)))
    if not local:
        mat = mat / np.sqrt(np.sum(mat ** 2, axis=0, keepdims=True))
    return mat


# TODO Check why they omit m_squared?
def _divide_by_msq(mat):
    # how msq looks for uni-variate moran
    spot_n = mat.shape[0]
    msq = (np.sum(mat ** 2, axis=0) / (spot_n - 1))
    return mat / msq


def _get_ordered_matrix(mat, pos, order):
    _indx = np.array([pos[x] for x in order])
    return mat[:, _indx].T


def _global_zscore_pvals(dist, global_r, positive_only):
    """

    Parameters
    ----------
    dist
        proximity weight matrix (spot_n x spot_n)
    global_r
        Array with
    positive_only: bool
        whether to mask negative correlation p-values

    Returns
    -------
        1D array with the size of global_r

    """
    dist = np.array(dist.todense())
    spot_n = dist.shape[0]

    # global distance variance as in spatialDM
    numerator = spot_n ** 2 * ((dist * dist).sum()) - \
                (2 * spot_n * (dist @ dist).sum()) + \
                (dist.sum() ** 2)
    denominator = spot_n ** 2 * (spot_n - 1) ** 2
    dist_var_sq = (numerator / denominator) ** (1 / 2)

    global_zscores = global_r / dist_var_sq

    if positive_only:
        global_zpvals = norm.sf(global_zscores)
    else:
        global_zpvals = norm.sf(np.abs(global_zscores)) * 2

    return global_zpvals


def _calculate_local_moransI(x_mat, y_mat, dist):
    """

    Parameters
    ----------
    x_mat
        2D array with x variables
    y_mat
        2D array with y variables
    dist
    
    Returns
    -------
    Returns 2D array of local Moran's I with shape(n_spot, xy_n)

    """
    local_x = x_mat * (dist @ y_mat)
    local_y = x_mat * (dist @ y_mat)
    local_r = (local_x + local_y).T

    return local_r


def _local_zscore_pvals(x_mat, y_mat, local_r, dist, positive_only):
    """

    Parameters
    ----------
    x_mat
        2D array with x variables
    y_mat
        2D array with y variables
    local_r
        2D array with Local Moran's I
    dist
        proximity weights
    positive_only
        Whether to mask negative correlations pvalue

    Returns
    -------
    2D array of p-values with shape(n_spot, xy_n)

    """
    spot_n = dist.shape[0]

    x_norm = np.apply_along_axis(norm.fit, axis=0, arr=x_mat)
    y_norm = np.apply_along_axis(norm.fit, axis=0, arr=y_mat)

    # get x,y std
    x_sigma, y_sigma = x_norm[1, :], y_norm[1, :]

    x_sigma = x_sigma * spot_n / (spot_n - 1)
    y_sigma = y_sigma * spot_n / (spot_n - 1)

    std = _get_local_var(x_sigma, y_sigma, dist, spot_n)
    local_zscores = local_r / std

    if positive_only:
        local_zpvals = norm.sf(local_zscores)
        pos_msk = ((x_mat > 0) + (y_mat > 0)).T  # mask?
        local_zpvals[~pos_msk] = 1
    else:
        local_zpvals = norm.sf(np.abs(local_zscores))

    return local_zpvals


def _get_local_var(x_sigma, y_sigma, dist, spot_n):
    """

    Parameters
    ----------
    x_sigma
        Standard deviations for each x (e.g. std of all ligands in the matrix)
    y_sigma
        Standard deviations for each y (e.g. std of all receptors in the matrix)
    dist
        proximity weight matrix
    spot_n
        number of spots/cells in the matrix

    Returns
    -------
    2D array of standard deviations with shape(n_spot, xy_n)

    """
    dist_sq = (np.array(dist.todense()) ** 2).sum(axis=1)

    n_weight = 2 * (spot_n - 1) ** 2 / spot_n ** 2
    sigma_prod = x_sigma * y_sigma
    core = n_weight * sigma_prod

    var = np.multiply.outer(dist_sq, core) + core
    std = var ** (1 / 2)

    return std.T


# initialize instance
_spatialdm = SpatialMethod(
    method_name="SpatialDM",
    key_cols=['ligand_complex', 'receptor_complex'],
    reference="Zhuoxuan, L.I., Wang, T., Liu, P. and Huang, Y., 2022. SpatialDM: Rapid "
              "identification of spatially co-expressed ligand-receptor reveals cell-cell "
              "communication patterns. bioRxiv. "
)

spatialdm = SpatialDM(_method=_spatialdm,
                      _complex_cols=['ligand_means', 'receptor_means'],
                      _obsm_keys=['proximity']
                      )





import numpy as np
from scanpy.datasets import pbmc68k_reduced
from  scipy.sparse import csr_matrix

from liana.method.sp._spatialdm import spatialdm, _global_zscore_pvals, _local_zscore_pvals, \
    _global_permutation_pvals, _calculate_local_moransI
from liana.method.sp._spatial_utils import _local_permutation_pvals

adata = pbmc68k_reduced()
proximity = np.zeros([adata.shape[0], adata.shape[0]])
np.fill_diagonal(proximity, 1)
adata.obsm['proximity'] = proximity


# toy test data
seed = 0
rng = np.random.default_rng(seed=seed)
dist = csr_matrix(rng.normal(size=(10, 10)))
x_mat = rng.normal(size=(10, 10))
y_mat = rng.normal(size=(10, 10))
n_perm = 100
positive_only = True
    

def test_global_permutation_pvals():
    global_truth = rng.normal(size=(10))

    pvals = _global_permutation_pvals(x_mat=x_mat,
                                      y_mat=y_mat,
                                      global_r=global_truth,
                                      seed=seed,  
                                      n_perm=n_perm,
                                      positive_only=positive_only,
                                      dist = dist)
    assert pvals.shape == (10, )
    assert pvals.sum().round(3)==4.65
    
    

def test_local_permutation_pvals():
    local_truth = rng.normal(size=(10, 10))
    positive_only = True

    pvals = _local_permutation_pvals(x_mat = x_mat,
                                     y_mat = y_mat,
                                     local_truth = local_truth,
                                     local_fun = _calculate_local_moransI,
                                     dist = dist,
                                     n_perm = n_perm,
                                     seed = seed,
                                     positive_only=positive_only)
    assert pvals.shape == (10, 10)
        


def test_global_zscore_pvals():
    global_truth = rng.normal(size=(10))
    pvals = _global_zscore_pvals(global_r=global_truth, dist=dist, positive_only=positive_only)
    assert pvals.shape == (10, )


def test_local_zscore_pvals():
    local_truth = rng.normal(size=(10, 10))
    pvals = _local_zscore_pvals(x_mat=x_mat, y_mat=y_mat, dist=dist, local_r=local_truth, positive_only=positive_only)
    assert pvals.shape == (10, 10)



def test_spatialdm():
    spatialdm(adata, use_raw=True)
    assert 'global_res' in adata.uns_keys()
    assert 'local_r' in adata.obsm_keys()
    assert 'local_pvals' in adata.obsm_keys()

    # test specific interaction
    global_res = adata.uns['global_res']
    interaction = global_res[global_res.interaction == 'S100A9&ITGB2']
    np.testing.assert_almost_equal(interaction['global_r'].values, 0.3100374)
    np.testing.assert_almost_equal(interaction['global_pvals'].values, 1.232729e-16)

    assert np.mean(adata.obsm['local_r']['MIF&CD74_CXCR4']) == 0.024113696644942034
    assert np.mean(adata.obsm['local_pvals']['TNFSF13B&TNFRSF13B']) == 0.9214972206911888


def test_spatialdm_permutation():
    spatialdm(adata, use_raw=True, pvalue_method="permutation")
    assert 'global_res' in adata.uns_keys()
    assert 'local_r' in adata.obsm_keys()
    assert 'local_pvals' in adata.obsm_keys()
    
    global_res = adata.uns['global_res']
    interaction = global_res[global_res.interaction == 'S100A9&ITGB2']
    
    np.testing.assert_almost_equal(interaction['global_r'].values, 0.3100374)
    np.testing.assert_almost_equal(interaction['global_pvals'].values, 0.0)
    
    assert np.mean(adata.obsm['local_r']['MIF&CD74_CXCR4']) == 0.024113696644942034
    assert np.mean(adata.obsm['local_pvals']['TNFSF13B&TNFRSF13B']) == 0.9874585714285714



from pandas import DataFrame


# TODO MethodMeta class to generalizable, SpatialMethod & SingleCellMethod should inherit and extend
class SpatialMethod:
    """
    A SpatialMethod Class
    """

    def __init__(self,
                 method_name: str,
                 key_cols: list,  # note that this is defined here but not in Method
                 reference: str
                 ):
        """
        Parameters
        ----------
        method_name
            Name of the Method
        key_cols
            columns which make every interaction unique (i.e. PK).
        reference
            Publication reference in Harvard style
        """
        self.method_name = method_name
        self.key_cols = key_cols
        self.reference = reference

    def describe(self):
        """Briefly described the method"""
        print(f"{ self.method_name } does XYZ")

    def reference(self):
        """Prints out reference in Harvard format"""
        print(self.reference)

    def get_meta(self):
        """Returns method metadata as pandas row"""
        meta = DataFrame([{"Method Name": self.method_name,
                           "Reference": self.reference
                           }])
        return meta



import numpy as np
import pandas as pd
import anndata
from pandas import DataFrame

from sklearn.neighbors import NearestNeighbors
from scipy.spatial.distance import pdist, squareform
from scipy.sparse import csr_matrix
from tqdm import tqdm


def get_spatial_proximity(adata: anndata.AnnData,
                          parameter,
                          family='gaussian',
                          cutoff=None,
                          n_neighbors=None,
                          bypass_diagonal=False,
                          inplace=True):
    """
    Generate spatial proximity weights using Euclidean distance.

    Parameters
    ----------
    adata
        `AnnData` object with spatial coordinates (in 'spatial') in `adata.obsm`.
    parameter
         Denotes signaling length (`l`)
    family
        Functions used to generate proximity weights. The following options are available:
        ['gaussian', 'spatialdm', 'exponential', 'linear']
    cutoff
        Vales below this cutoff will be set to 0
    n_neighbors
        Find k nearest neighbours, use it as a proximity mask. In other words,
        only the proximity of the nearest neighbours is kept as calculated
        by the specified radial basis function, the remainder are set to 0.
    bypass_diagonal
        Logical, sets proximity diagonal to 0 if true.
    inplace
        If true return `DataFrame` with results, else assign to `.obsm`.

    Returns
    -------
    If ``inplace = False``, returns an `np.array` with spatial proximity weights.
    Otherwise, modifies the ``adata`` object with the following key:
        - :attr:`anndata.AnnData.obsm` ``['proximity']`` with the aforementioned array
    """

    families = ['gaussian', 'spatialdm', 'exponential', 'linear']
    if family not in families:
        raise AssertionError(f"{family} must be a member of {families}")

    if (cutoff is None) & (n_neighbors is None):
        raise ValueError("`cutoff` or `n_neighbors` must be provided!")

    assert 'spatial' in adata.obsm

    coordinates = pd.DataFrame(adata.obsm['spatial'],
                               index=adata.obs_names,
                               columns=['x', 'y'])

    proximity = pdist(coordinates, 'euclidean')
    proximity = squareform(proximity)

    # prevent overflow
    proximity = np.array(proximity, dtype=np.float64)
    parameter = np.array(parameter, dtype=np.float64)

    if family == 'gaussian':
        proximity = np.exp(-(proximity ** 2.0) / (2.0 * parameter ** 2.0))
    elif family == 'misty_rbf':
        proximity = np.exp(-(proximity ** 2.0) / (parameter ** 2.0))
    elif family == 'exponential':
        proximity = np.exp(-proximity / parameter)
    elif family == 'linear':
        proximity = 1 - proximity / parameter
        proximity[proximity < 0] = 0

    if bypass_diagonal:
        np.fill_diagonal(proximity, 0)

    if cutoff is not None:
        proximity[proximity < cutoff] = 0
    if n_neighbors is not None:
        nn = NearestNeighbors(n_neighbors=n_neighbors).fit(proximity)
        knn = nn.kneighbors_graph(proximity).toarray()
        proximity = proximity * knn  # knn works as mask

    spot_n = proximity.shape[0]
    assert spot_n == adata.shape[0]

    # speed up
    if spot_n > 1000:
        proximity = proximity.astype(np.float16)

    proximity = csr_matrix(proximity)

    adata.obsm['proximity'] = proximity
    return None if inplace else proximity


def _local_to_dataframe(idx, columns, array):
    return DataFrame(array, index=idx, columns=columns)



def _local_permutation_pvals(x_mat, y_mat, dist, local_truth, local_fun,n_perm, seed, positive_only, **kwargs):
    """
    Calculate local pvalues for a given local score function.

    Parameters
    ----------
    x_mat
        2D array with x variables
    y_mat
        2D array with y variables
    local_truth
        2D array with non-permuted local scores/co-expressions
    dist
        proximity weights
    n_perm
        number of permutations
    seed
        Reproducibility seed
    positive_only
        Whether to mask negative correlations pvalue

    Returns
    -------
    2D array with shape(n_spot, xy_n)

    """
    rng = np.random.default_rng(seed)
    assert isinstance(dist, csr_matrix)

    xy_n = local_truth.shape[0]
    spot_n = local_truth.shape[1]
    
    # permutation cubes to be populated
    local_pvals = np.zeros((xy_n, spot_n))
    
    # shuffle the matrix
    for i in tqdm(range(n_perm)):
        _idx = rng.permutation(spot_n)
        perm_score = local_fun(x_mat = x_mat[_idx, :], y_mat=y_mat, dist=dist, **kwargs)
        if positive_only:
            local_pvals += np.array(perm_score >= local_truth, dtype=int)
        else:
            local_pvals += (np.array(np.abs(perm_score) >= np.abs(local_truth), dtype=int))

    local_pvals = local_pvals / n_perm

    if positive_only:  # TODO change to directed mask (both, negative, positive)
        # only keep positive pvals where either x or y is positive
        pos_msk = ((x_mat > 0) + (y_mat > 0)).T
        local_pvals[~pos_msk] = 1

    return local_pvals



import numpy as np
import pandas as pd

from liana.resource import select_resource
from liana.method._pipe_utils._reassemble_complexes import explode_complexes
from liana.method._pipe_utils import prep_check_adata, filter_resource, assert_covered, filter_reassemble_complexes
from liana.utils._utils import _get_props


def _global_lr_pipe(adata,
                    resource,
                    resource_name,
                    expr_prop,
                    use_raw,
                    layer,
                    verbose,
                    _key_cols,
                    _complex_cols,
                    _obms_keys,
                    ):
    """

    Parameters
    ----------
    adata
        The annotated data matrix of shape `n_obs` × `n_vars`. Rows correspond to cells and
    resource
        Ligand-receptor resource as a pandas dataframe with columns: ligand, receptor
    resource_name
        Name of the resource
    expr_prop
        Proportion of expression of the ligand and receptor in a cell for the interaction to be considered
    layer
        Layer to use for the analysis. If None, use raw or X
    verbose
        Verbosity
    _obms_keys
        Keys of the adata.obsm to use for the analysis
    _key_cols
        Columns to use as keys for the resource
    _complex_cols
        Columns to use as keys for the complexes

    Returns
    -------
    Returns adata, lr_res, ligand_pos, receptor_pos
    

    """
    # prep adata
    adata = prep_check_adata(adata=adata,
                             use_raw=use_raw,
                             layer=layer,
                             verbose=verbose,
                             groupby=None,
                             min_cells=None,
                             obsm_keys=_obms_keys
                             )

    # select & process resource
    if resource is None:
        resource = select_resource(resource_name.lower())
    resource = explode_complexes(resource)
    resource = filter_resource(resource, adata.var_names)

    # get entities
    entities = np.union1d(np.unique(resource["ligand"]),
                          np.unique(resource["receptor"]))

    # Check overlap between resource and adata  TODO check if this works
    assert_covered(entities, adata.var_names, verbose=verbose)

    # Filter to only include the relevant features
    adata = adata[:, np.intersect1d(entities, adata.var.index)]

    # global_stats
    global_stats = pd.DataFrame({'means': adata.X.mean(axis=0).A.flatten(),
                                 'props': _get_props(adata.X)},
                                index=adata.var_names).reset_index().rename(
        columns={'index': 'gene'})

    # join global stats to LRs from resource
    lr_res = resource.merge(_rename_means(global_stats, entity='ligand')).merge(
        _rename_means(global_stats, entity='receptor'))

    # get lr_res /w relevant x,y (lig, rec) and filter acc to expr_prop
    lr_res = filter_reassemble_complexes(lr_res=lr_res,
                                         expr_prop=expr_prop,
                                         _key_cols=_key_cols,
                                         complex_cols=_complex_cols
                                         )

    # assign the positions of x, y to the adata
    ligand_pos = {entity: np.where(adata.var_names == entity)[0][0] for entity in
                  lr_res['ligand']}
    receptor_pos = {entity: np.where(adata.var_names == entity)[0][0] for entity in
                    lr_res['receptor']}

    return adata, lr_res, ligand_pos, receptor_pos


def _rename_means(lr_stats, entity):
    df = lr_stats.copy()
    df.columns = df.columns.map(lambda x: entity + '_' + str(x) if x != 'gene' else 'gene')
    return df.rename(columns={'gene': entity})



"""
Preprocessing functions.
Functions to preprocess the anndata object prior to running any method.
"""
from __future__ import annotations

import numpy as np
from anndata import AnnData
from typing import Optional
from pandas import DataFrame, Index
import scanpy as sc
from scipy.sparse import csr_matrix, isspmatrix_csr


def assert_covered(
        subset,
        superset,
        subset_name: str = "resource",
        superset_name: str = "var_names",
        prop_missing_allowed: float = 0.98,
        verbose: bool = False) -> None:
    """
    Assert if elements are covered at a decent proportion

    Parameters
    ----------
    subset
        Subset of elements
    superset
        The superset of elements
    subset_name
        Name of the subset
    superset_name
        Name of the superset
    prop_missing_allowed
        Allowed proportion of missing/mismatched elements in the subset
    verbose
        Verbosity flag

    Returns
    -------
    None
    """

    subset = np.asarray(subset)
    is_missing = ~np.isin(subset, superset)
    prop_missing = np.sum(is_missing) / len(subset)
    x_missing = ", ".join([x for x in subset[is_missing]])

    if prop_missing > prop_missing_allowed:
        msg = (
            f"Please check if appropriate organism/ID type was provided! "
            f"Allowed proportion ({prop_missing_allowed}) of missing "
            f"{subset_name} elements exceeded ({prop_missing:.2f}). "
            f"Too few features from the resource were found in the data."
        )
        raise ValueError(msg + f" [{x_missing}] missing from {superset_name}")

    if verbose & (prop_missing > 0):
        print(f"{prop_missing:.2f} of entities in the resource are missing from the data.")


def prep_check_adata(adata: AnnData,
                     groupby: (str | None),
                     min_cells: (int | None),
                     use_raw: Optional[bool] = False,
                     layer: Optional[str] = None,
                     obsm_keys=None,
                     verbose: Optional[bool] = False) -> AnnData:
    """
    Check if the anndata object is in the correct format and preprocess

    Parameters
    ----------
    adata
        Un-formatted Anndata.
    groupby
        column to groupby. None if the ligand-receptor pipe
        calling this function does not rely on cell labels.
        For example, if ligand-receptor stats are needed
        for the whole sample (global).
    min_cells
        minimum cells per cell identity. None if groupby is not passed.
    use_raw
        Use raw attribute of adata if present.
    layer
        Indicate whether to use any layer.
    obsm_keys
        Indicate whether to keep obsm with spatial info or to discard.
        By default, False and discarded.
    verbose
        Verbosity flag.

    Returns
    -------
    Anndata object to be used downstream
    """
    # simplify adata

    X = _choose_mtx_rep(adata=adata, use_raw=use_raw,
                        layer=layer, verbose=verbose)

    if use_raw & (layer is None):
        var = adata.raw.var.copy()
    else:
        var = adata.var.copy()

    if obsm_keys:
        obsm = {key: adata.obsm[key] for key in obsm_keys}
    else:
        obsm = None

    adata = sc.AnnData(X=X,
                       obs=adata.obs.copy(),
                       dtype="float32",
                       var=var,
                       obsm=obsm
                       )

    # convert to sparse csr matrix
    if not isspmatrix_csr(adata.X):
        if verbose:
            print("Converting mat to CSR format")
        adata.X = csr_matrix(adata.X)

    # Check for empty features
    msk_features = np.sum(adata.X, axis=0).A1 == 0
    n_empty_features = np.sum(msk_features)
    if n_empty_features > 0:
        if verbose:
            print("{0} features of mat are empty, they will be removed.".format(
                n_empty_features))
        adata = adata[:, ~msk_features]

    # Check for empty samples
    msk_samples = np.sum(adata.X, axis=1).A1 == 0
    n_empty_samples = np.sum(msk_samples)
    if n_empty_samples > 0:
        raise ValueError("{0} cells are empty, please remove those!")

    # Check if log-norm
    _sum = np.sum(adata.X.data[0:100])
    if _sum == np.floor(_sum):
        if verbose:
            print("Make sure that normalized counts are passed!")

    # Check for non-finite values
    if np.any(~np.isfinite(adata.X.data)):
        raise ValueError(
            """mat contains non finite values (nan or inf), please set them 
            to 0 or remove them.""")

    # Define idents col name
    if groupby is not None:
        if groupby not in adata.obs.columns:
            raise AssertionError(f"`{groupby}` not found in `adata.obs.columns`.")
        adata.obs.loc[:, 'label'] = adata.obs[groupby]

        # Remove any cell types below X number of cells per cell type
        count_cells = adata.obs.groupby(groupby)[groupby].size().reset_index(name='count').copy()
        count_cells['keep'] = count_cells['count'] >= min_cells

        if not all(count_cells.keep):
            lowly_abundant_idents = list(count_cells[~count_cells.keep][groupby])
            # remove lowly abundant identities
            msk = ~np.isin(adata.obs[[groupby]], lowly_abundant_idents)
            adata = adata[msk]
            if verbose:
                print("The following cell identities were excluded: {0}".format(
                    ", ".join(lowly_abundant_idents)))

    # Re-order adata vars alphabetically
    adata = adata[:, np.sort(adata.var_names)]

    # Remove underscores from gene names
    adata.var_names = format_vars(adata.var_names)

    return adata


# Helper function to replace a substring in string and append to list
def _append_replace(x: str, l: list):
    l.append(x)
    return x.replace('_', '')


# format variable names
def format_vars(var_names, verbose=False) -> list:
    """
    Format Variable names
    
    Parameters
    ----------
    var_names
        list of variable names (e.g. adata.var_names)
    verbose
        Verbosity flag

    Returns
    -------
        Formatted Variable names list

    """
    changed = []
    var_names = [_append_replace(x, changed) if ('_' in x) else x for x in var_names]
    changed = ' ,'.join(changed)
    if verbose & (len(changed) > 0):
        print(f"Replace underscores (_) with blank in {changed}", )
    return var_names


def filter_resource(resource: DataFrame, var_names: Index) -> DataFrame:
    """
    Filter interactions for which vars are not present.

    Note that here I remove any interaction that /w genes that are not found
    in the dataset. Note that this is not necessarily the case in liana-r.
    There, I assign the expression of those with missing subunits to 0, while
    those without any subunit present are implicitly filtered.

    Parameters
    ---------
    resource
        Resource with 'ligand' and 'receptor' columns
    var_names
        Relevant variables - i.e. the variables to be used downstream

    Returns
    ------
    A filtered resource dataframe
    """
    # Remove those without any subunit
    resource = resource[(np.isin(resource.ligand, var_names)) &
                        (np.isin(resource.receptor, var_names))]

    # Only keep interactions /w complexes for which all subunits are present
    missing_comps = resource[resource.interaction.str.contains('_')].copy()
    missing_comps['all_units'] = \
        missing_comps['ligand_complex'] + '_' + missing_comps[
            'receptor_complex']

    # Get those not with all subunits
    missing_comps = missing_comps[np.logical_not(
        [all([x in var_names for x in entity.split('_')])
         for entity in missing_comps.all_units]
    )]
    # Filter them
    return resource[~resource.interaction.isin(missing_comps.interaction)]


def _choose_mtx_rep(adata, use_raw=False, layer=None, verbose=False) -> csr_matrix:
    """
    Choose matrix (adapted from scanpy)
    
    Parameters
    ----------
    adata
        Unformatted Anndata.
    use_raw
        Use raw attribute of adata if present.
    layer
        Indicate whether to use any layer.
    
    Returns
    -------
        The matrix to be used by liana-py.
    """

    is_layer = layer is not None
    if is_layer & use_raw:
        raise ValueError("Cannot specify `layer` and have `use_raw=True`.")
    if is_layer:
        if verbose:
            print(f"Using the `{layer}` layer!")
        return adata.layers[layer]
    elif use_raw:
        if adata.raw is None:
            raise ValueError("`.raw` is not initialized!")
        if verbose:
            print("Using `.raw`!")
        return adata.raw.X
    else:
        if verbose:
            print("Using `.X`!")
        return adata.X



import numpy as np
from scanpy.datasets import pbmc68k_reduced
from liana.method import get_spatial_proximity

# TODO move all of this repetitive code to a single function in testing
adata = pbmc68k_reduced()
proximity = np.zeros([adata.shape[0], adata.shape[0]])

rng = np.random.default_rng(seed=1337)
x = rng.integers(low=0, high=5000, size=adata.shape[0])
y = rng.integers(low=0, high=5000, size=adata.shape[0])
adata.obsm['spatial'] = np.array([x, y]).T
get_spatial_proximity(adata, parameter=100, cutoff=0.1)


def test_get_spatial_proximity():
    get_spatial_proximity(adata=adata, parameter=200, bypass_diagonal=False, cutoff=0.2)




from ..multi import to_tensor_c2c
from ..testing import sample_lrs
import cell2cell as c2c


def test_to_tensor_c2c():
    """Test to_tensor_c2c."""
    liana_res = sample_lrs(by_sample=True)
    liana_res = liana_res.drop_duplicates(['source', 'target', 'ligand_complex', 'receptor_complex'])
    
    liana_dict = to_tensor_c2c(liana_res=liana_res,
                               sample_key='sample',
                               score_key='specificity_rank',
                               return_dict=True
                               )
    assert isinstance(liana_dict, dict)
    
    tensor = to_tensor_c2c(liana_res=liana_res,
                           sample_key='sample',
                           score_key='specificity_rank')
    assert isinstance(tensor, c2c.tensor.tensor.PreBuiltTensor)
    assert tensor.sparsity_fraction()==0.0



from __future__ import annotations

import pandas as pd
import numpy as np
from anndata import AnnData


def _get_props(X_mask):
    return X_mask.getnnz(axis=0) / X_mask.shape[0]


def obsm_to_adata(adata, obsm_key):
    """
    Extracts activities as AnnData object.
    From an AnnData object with source activities stored in `.obsm`, generates a new AnnData object with activities in `X`.
    This allows to reuse many scanpy processing and visualization functions.

    Parameters
    ----------
    adata : AnnData
        Annotated data matrix with activities stored in .obsm.
    obsm_key
        `.osbm` key to extract.
    Returns
    -------
    acts : AnnData
        New AnnData object with activities in X.
    """

    obs = adata.obs
    var = pd.DataFrame(index=adata.obsm[obsm_key].columns)
    uns = adata.uns
    obsm = adata.obsm
    X = np.array(adata.obsm[obsm_key], dtype=np.float32)

    return AnnData(X=X, obs=obs, var=var, uns=uns, obsm=obsm)



from plotnine import ggplot, aes, geom_point, theme_minimal, labs
import anndata
import pandas as pd


def proximity_plot(adata: anndata.AnnData, idx: int, return_fig: bool = True):
    """
    Plot spatial proximity weights.

    Parameters
    ----------
    adata
        `AnnData` object with `proximity` (spatial proximity weights) in `adata.obsm`.
    idx
        Spot/cell index
    return_fig
        `bool` whether to return the fig object, `False` only plots

    Returns
    -------
    A `plotnine.ggplot` instance

    """

    assert 'proximity' in adata.obsm_keys()
    assert 'spatial' in adata.obsm_keys()

    coordinates = pd.DataFrame(adata.obsm['spatial'],
                               index=adata.obs_names,
                               columns=['x', 'y']).copy()
    coordinates['proximity'] = adata.obsm['proximity'][:, idx].A

    p = (ggplot(coordinates.sort_values('proximity', ascending=True),
                aes(x='x', y='y', colour='proximity'))
         + geom_point(size=2.7, shape='8')
         + theme_minimal()
         + labs(colour='Proximity', y='y Coordinate', x='x Coordinate')
         )

    if return_fig:
        return p

    p.draw()
